unit universe;

interface

uses ExtCtrls, vector,  Classes, Controls, Graphics, StdCtrls;

type
  PControl=^TControl;
  PStrings=^TStrings;
  PShape=^TShape;
  ptrk=^trk;
  trk=record
      next: ptrk;
      pos: vektor;
  end;

{Jak budou vypadat planety}
PPlanet=^Tplanet;
TPlanet=class(TShape)
  public
      name: string;
      procedure SetRadius(ir: double);
      procedure SetMass(im: double);
      function GetRadius: double;
      function GetMass: double;
      procedure SetPosition(ix: double; iy:double; iz: double); overload; virtual;
      procedure SetVelocity(what: vektor); virtual;
      procedure SetPosition(where: vektor); overload; virtual;
      procedure GetPosition(var ix: double; var iy: double; var iz: double); overload;
      procedure GetPosition(var ix: variant; var iy: variant; var iz: variant); overload;
      procedure GetPosition(var where: vektor); overload;
      procedure GetVelocity(var what: vektor); overload;
      procedure GetVelocity(var ix: variant; var iy: variant; var iz: variant); overload;
      procedure AddTime(time: double);
      procedure Iteraguj(skym: PPlanet);
      constructor create(owner: PControl; PlanetName: string);
      destructor Destroy; override;
      procedure aktualize;
  private
      pos: vektor;
      Track: ptrk;
      r: double;
      m: double;
      f: vektor;
      v: vektor;
      owner: PControl;
  end;


{par pomocnych veci}

pspoj=^spoj;
spoj=record
    next: pspoj;
    planeta: PPlanet;
end;

{a jak bude vypadat vesmir}
TUniverse=class(TPanel)
  public
    Sit: PPlanet;
    Look: PPlanet;
    Time: double;
    Obraz: TImage;
    constants: record
      h: double;
      TrackLenght: longint;
      v: double;
      t: double;
      m: double;
      f: double;
      l: double;
    end;
    constructor Create(owner: Tcomponent); override;
    procedure AddPlanet(Which: PPlanet);
    procedure DestroyPlanet(Which: string);
    procedure AddTime(time: double);
    procedure Refresh;
    function GetPlanet(Name: string): PPlanet;
    procedure GetPlanets(kam: PStrings);
    function ZDeform(Z: double): double;
  private
    Planets: pspoj;
end;

PUniverse=^TUniverse;


implementation

uses main,
  Windows, Messages, SysUtils, Variants, Forms,
  Dialogs, Buttons;

{implementace vesmiru}


{konstruktor}
constructor TUniverse.Create(owner: TComponent);
begin
  inherited Create(owner);
  Obraz:=TImage.Create(self);
  self.InsertControl(obraz);
  obraz.Top:=1;
  obraz.Left:=1;
  obraz.SendToBack;
  self.BevelOuter:=bvLowered;
  self.SendToBack;
  self.Planets:=NIL;
  self.Sit:=NIL;
  self.Look:=NIL;
  self.Color:=clBlack;
  { konstanty a jednotky}
  self.constants.h:=6.67/power(10,11);  {gravitacni konstanta}
  self.constants.t:=10*24*60*60;        {jednotkou casu je 10 dni}
  self.constants.m:=power(10,22);       {jednotkou hmotnosti je 10^22kg}
  self.constants.l:=power(10,8);        {jednotkou vzdalenosti budou statisice kilometru}
  self.constants.v:=1;                  {jednotkou rychlosti budou m/s}
  self.constants.tracklenght:=0;        {zadna stopa}
  self.constants.f:=0;                  {nezobrazovat pusobici sily}
  self.Time:=0;
end;

{pridani planety}
procedure TUniverse.AddPlanet(Which: PPlanet);
var a: pspoj;
begin
  inherited InsertControl(which^);
   new(a);
  a.next:=self.Planets;
  a.planeta:=which;
  self.Planets:=a;
end;

{vrati cislo pro zmenu souradnic pri rostouci vzdalenosti}
function TUniverse.ZDeform(Z: double): double;
begin
    zdeform:=(sqrt(abs(z)/self.constants.l)+1)*self.constants.l;
end;

{ prekresleni vesmiru }
procedure TUniverse.Refresh;
var a,b: pspoj;
    fil: TRect;
    alpha, beta, temp: double;
    tmp, vtmp: vektor;
    spoj: ptrk;

begin
  {potrebna uprava souradnic}
  tmp.null;
  b:=self.Planets;
  if (self.Sit<>NIL) then     { pokud jsem na nejaky planete, tak prepocitam souradnice}
  begin
    tmp.z:=self.sit^.r;
    tmp:=tmp.add(self.sit^.pos);
    while b<>NIL do
      begin
        b^.planeta^.pos:=(b^.planeta^.pos.add(tmp.negative));
        if b^.planeta^.Track<>NIL then
          b^.planeta^.Track^.Pos:=(b^.planeta^.Track^.pos.add(tmp.negative));
        b:=b^.next;
      end;
  end;
  b:=self.Planets;
  if (self.Look<>NIL) then  { pokud se na neco divam, budu muset orotovat vesmir. }
  begin
    alpha:=self.look.pos.GetZXAngle;
    beta:=self.look.pos.GetZYAngle;
    while (b<>NIL) and ((beta<>0) or (alpha<>0))do
    begin
      if (alpha<>0) then    {vlastni rotace podle y}
      begin
        tmp:=b^.planeta^.pos;
        vtmp:=tmp.add(b^.planeta^.v);
        {rotace planety}
        b^.planeta^.pos.z:=cos(tmp.GetZXAngle-alpha)*sqrt(tmp.z*tmp.z+tmp.x*tmp.x);
        b^.planeta^.pos.x:=sin(tmp.GetZXAngle-alpha)*sqrt(tmp.z*tmp.z+tmp.x*tmp.x);
        {zmena rychlosti}
        b^.planeta^.v.z:=cos(vtmp.GetZXAngle-alpha)*sqrt(vtmp.z*vtmp.z+vtmp.x*vtmp.x)-b^.planeta^.pos.z;
        b^.planeta^.v.x:=sin(vtmp.GetZXAngle-alpha)*sqrt(vtmp.z*vtmp.z+vtmp.x*vtmp.x)-b^.planeta^.pos.x;
        {zmena stopy}
        if b^.planeta^.Track<>NIL then
        begin
          spoj:=b^.planeta^.Track^.next;
          vtmp:=b^.planeta^.Track^.pos;
          b^.planeta^.Track^.pos.z:=cos(vtmp.GetZXAngle-alpha)*sqrt(vtmp.z*vtmp.z+vtmp.x*vtmp.x);
          b^.planeta^.Track^.pos.x:=sin(vtmp.GetZXAngle-alpha)*sqrt(vtmp.z*vtmp.z+vtmp.x*vtmp.x);
          tmp:=b^.planeta^.Track^.pos;
          while spoj<>NIL do
          begin
            vtmp:=vtmp.add(spoj.pos);
            temp:=cos(vtmp.GetZXAngle-alpha)*sqrt(vtmp.z*vtmp.z+vtmp.x*vtmp.x);
            spoj^.pos.z:=temp-tmp.z;
            tmp.z:=temp;
            temp:=sin(vtmp.GetZXAngle-alpha)*sqrt(vtmp.z*vtmp.z+vtmp.x*vtmp.x);
            spoj^.pos.x:=temp-tmp.x;
            tmp.x:=temp;
            spoj:=spoj^.next;
          end;
        end;
      end;
      if (beta<>0) then    {vlastni rotace podle x}
      begin
        tmp:=b^.planeta^.pos;
        vtmp:=tmp.add(b^.planeta^.v);
        {rotace planety}
        b^.planeta^.pos.z:=cos(tmp.GetZYAngle-beta)*sqrt(tmp.z*tmp.z+tmp.y*tmp.y);
        b^.planeta^.pos.y:=sin(tmp.GetZYAngle-beta)*sqrt(tmp.z*tmp.z+tmp.y*tmp.y);
        {zmena rychlosti}
        b^.planeta^.v.z:=cos(vtmp.GetZYAngle-beta)*sqrt(vtmp.z*vtmp.z+vtmp.y*vtmp.y)-b^.planeta^.pos.z;
        b^.planeta^.v.y:=sin(vtmp.GetZYAngle-beta)*sqrt(vtmp.z*vtmp.z+vtmp.y*vtmp.y)-b^.planeta^.pos.y;
        {zmena stopy}
        if b^.planeta^.Track<>NIL then
        begin
          spoj:=b^.planeta^.Track^.next;
          vtmp:=b^.planeta^.Track^.pos;
          b^.planeta^.Track^.pos.z:=cos(vtmp.GetZYAngle-alpha)*sqrt(vtmp.z*vtmp.z+vtmp.y*vtmp.y);
          b^.planeta^.Track^.pos.y:=sin(vtmp.GetZYAngle-alpha)*sqrt(vtmp.z*vtmp.z+vtmp.y*vtmp.y);
          tmp:=b^.planeta^.Track^.pos;
          while spoj<>NIL do
          begin
            vtmp:=vtmp.add(spoj.pos);
            temp:=cos(vtmp.GetZYAngle-beta)*sqrt(vtmp.z*vtmp.z+vtmp.y*vtmp.y);
            spoj^.pos.z:=temp-tmp.z;
            tmp.z:=temp;
            temp:=sin(vtmp.GetZYAngle-beta)*sqrt(vtmp.z*vtmp.z+vtmp.y*vtmp.y);
            spoj^.pos.y:=temp-tmp.y;
            tmp.y:=temp;
            spoj:=spoj^.next;
          end;
        end;
      end;
      b:=b^.next;
    end;
  end;

  {prekresleni}
  with self.obraz.canvas do
  begin
    fil.Left:=0;
    fil.Top:=0;
    fil.Right:=Width;
    fil.Bottom:=height;
    brush.Color:=clBlack;
    pen.Color:=clBlack;
    self.Obraz.Canvas.FillRect(fil);
    pen.Color:=clWhite;
    moveto(10,8);
    lineto(10,16);
    moveto(10,12);
    lineto(110,12);
    moveto(110,8);
    lineto(110,16);
    pen.Color:=clBlack;
  end;
  a:=self.Planets;
  while a<>NIL do
  begin
    a.planeta^.aktualize;
    a:=a.next;
  end;

end;


{ zniceni planety}
procedure TUniverse.DestroyPlanet(which: string);
var a, b: pspoj;
begin
  b:=NIL;
  a:=self.Planets;
  while (a<>NIL) and (a^.planeta^.name<>which) do
  begin
    b:=a;
    a:=a^.next;
  end;
  if (b<>NIL) and (a<>NIL) then
    begin
      b^.next:=a^.next;
    end else
      if (b=NIL) and (a<>NIL) then self.Planets:=a^.next;
  if a<>NIL then
  begin
      a^.planeta^.Destroy;
      dispose(a^.planeta);
      dispose(a);
  end;
end;

{ starnuti vesmiru }
procedure TUniverse.AddTime(time: double);
var a,b: Pspoj;
begin
  a:=self.planets;
  while a<>NIL do             {vynulovani pusobicich sil}
  begin
    a^.planeta^.f.null;
    a:=a^.next;
  end;
  a:=self.planets;
  while a<>NIL do             {iterakce planet vuci sobe navzajem}
  begin
    b:=a^.next;
    while b<>NIL do
      begin
        a^.planeta^.Iteraguj(b.planeta);
        b:=b^.next;
      end;
    a^.planeta.AddTime(time);
    a:=a^.next;
  end;
  self.Time:=self.Time+time;
end;

{ najde planetu }
function TUniverse.GetPlanet(Name: string): PPlanet;
var last: pspoj;
begin
  last:=self.Planets;
  while (last<>NIL) and (last^.planeta^.Name <> Name) do
    last:=last^.next;
  if last<>NIL then
    begin
      GetPlanet:=last^.planeta;
    end else
      GetPlanet:=NIL;
end;


{ zapise do strings jmena vsech planet}
procedure TUniverse.GetPlanets(kam: PStrings);
var last: pspoj;
begin
  last:=self.Planets;
  while (last<>NIL) do
  begin
    kam^.Add(last^.planeta^.name);
    last:=last^.next;
  end;
end;



{implementace planet}



{ konstruktor }
constructor Tplanet.create(Owner: PControl; PlanetName: string);
begin
  inherited Create(owner^);
  self.Shape:=stCircle;
  self.name:=PlanetName;
  self.owner:=owner;
  self.Pen.Color:=clBlack;
  self.v.null;
  self.Track:=NIL;
  self.f.null;
  self.OnClick:=Form1.SetActivePlanet;
end;

{ destruktor }
destructor TPlanet.Destroy;
var a,b: ptrk;
begin
  a:=self.track;
  while a<>NIL do
  begin
    b:=a;
    a:=a^.next;
    Dispose(b);
  end;
  inherited Destroy;
end;

{ nastaveni a zjisteni hmotnosti }
procedure TPlanet.setmass(im: double);
begin
  self.m:=im*(owner^ as TUniverse).constants.m;
end;

function TPlanet.getmass:double;
begin
  getmass:=self.m/(owner^ as TUniverse).constants.m;
end;


{ nastaveni a zjisteni Radiusu }
procedure TPlanet.setRadius(ir: double);
begin
  self.r:=ir*(owner^ as TUniverse).constants.l;
  self.aktualize;
end;

function TPlanet.getRadius: double;
begin
  getradius:=self.r/(owner^ as TUniverse).constants.l;
end;

{ vykresli se }
procedure TPlanet.aktualize;
var a,b: ptrk;
    len: integer;
    tmp: vektor;
    midx,midy: real;
begin
if self.pos.z>=0 then
  begin
    {vykresleni planety}
    self.Height:=round(max(self.r*2/(owner^ as TUniverse).zdeform(self.pos.z),1))+2;
    self.Width:=round(max(self.r*2/(owner^ as TUniverse).zdeform(self.pos.z),1))+2;
    self.Left:=round(owner^.Width/2+self.pos.x/(owner^ as TUniverse).zdeform(self.pos.z)-self.width/2);
    self.Top:=round(owner^.Height/2-self.pos.y/(owner^ as TUniverse).zdeform(self.pos.z)-self.height/2);
    self.Visible:=true;
  end else begin
    self.visible:=false;
  end;

    midx:=(owner^ as TUniverse).Obraz.Width/2;
    midy:=(owner^ as TUniverse).Obraz.Height/2;

    {vykresleni pusobici sily}
    if (owner^ as TUniverse).constants.f<>0 then
    begin
      (owner^ as TUniverse).Obraz.Canvas.Pen.Color:=clMaroon;
      (owner^ as TUniverse).Obraz.Canvas.MoveTo(round(midx+self.pos.x/(owner^ as TUniverse).zdeform(self.pos.z)), round(midy-self.pos.y/(owner^ as TUniverse).zdeform(self.pos.z)));
      tmp:=self.pos.add(self.f.multipliedby((owner^ as TUniverse).constants.f));
      (owner^ as TUniverse).Obraz.Canvas.LineTo(round(midx+tmp.x/(owner^ as TUniverse).zdeform(tmp.z)), round(midy-tmp.y/(owner^ as TUniverse).zdeform(tmp.z)));
    end;

    {pridani stopy}
    new(a);
    if self.Track<>NIL then
      self.Track^.pos:=self.Track^.pos.add(self.pos.negative);
    a^.next:=self.Track;
    a^.pos:=self.pos;
    self.Track:=a;

if (owner^ as Tuniverse).constants.TrackLenght>0 then
  begin
    {vykresleni stopy}
    (owner^ as TUniverse).Obraz.canvas.Pen.Color:=clGray;
    midx:=(owner^ as TUniverse).Obraz.Width/2;
    midy:=(owner^ as TUniverse).Obraz.Height/2;
    tmp:=self.pos;
    (owner^ as TUniverse).Obraz.Canvas.MoveTo(round(midx+tmp.x/(owner^ as TUniverse).zdeform(self.pos.z)), round(midy-tmp.y/(owner^ as TUniverse).zdeform(tmp.z)));
    a:=self.Track^.next;
    len:=1;
    while (len<=(owner^ as TUniverse).constants.TrackLenght) and (a<>NIL) do
    begin
      tmp:=tmp.add(a^.pos);
      if tmp.z>=0 then
        (owner^ as TUniverse).Obraz.Canvas.LineTo(round(midx+tmp.x/(owner^ as TUniverse).zdeform(tmp.z)), round(midy-tmp.y/(owner^ as TUniverse).zdeform(tmp.z)));
      (owner^ as TUniverse).Obraz.Canvas.MoveTo(round(midx+tmp.x/(owner^ as TUniverse).zdeform(tmp.z)), round(midy-tmp.y/(owner^ as TUniverse).zdeform(tmp.z)));
      a:=a^.next;
      inc(len);
    end;
  end;

    {smazani prebytecne stopy}
    if a<>NIL then
    begin
      if (owner^ as TUniverse).constants.TrackLenght=0 then
        self.Track:=NIL;
      b:=a^.next;
      a^.next:=NIL;
      a:=b;
      while a<>NIL do
      begin
        b:=a;
        a:=a^.next;
        Dispose(b);
      end;
    end;
end;


{ nastaveni pozice }
procedure TPlanet.setposition(ix: double; iy:double; iz: double);
begin
  self.pos.x:=ix*(owner^ as TUniverse).constants.l;
  self.pos.y:=iy*(owner^ as TUniverse).constants.l;
  self.pos.z:=iz*(owner^ as TUniverse).constants.l;
end;

procedure TPlanet.setposition(where: vektor);
begin
  self.setposition(where.x, where.y, where.z);
end;

{ precteni pozice }
procedure TPlanet.getposition(var ix: double; var iy: double; var iz: double);
begin
  ix:=self.pos.x/(owner^ as TUniverse).constants.l;
  iy:=self.pos.y/(owner^ as TUniverse).constants.l;
  iz:=self.pos.z/(owner^ as TUniverse).constants.l;
end;

procedure TPlanet.getposition(var ix: variant; var iy: variant; var iz: variant);
var xx,yy,zz: double;
begin
  self.GetPosition(xx,yy,zz);
  ix:=xx;
  iy:=yy;
  iz:=zz;
end;

procedure TPlanet.getposition(var where: vektor);
begin
  self.getposition(where.x, where.y, where.z);
end;


{nastaveni rychlosti}
procedure TPlanet.setvelocity(what: vektor);
var tmp: vektor;
begin
  tmp.null;
  if (self.owner^ as TUniverse).Sit<>NIL then
    tmp:=(self.owner^ as TUniverse).Sit^.v;
  self.v.x:=what.x*(self.owner^ as TUniverse).constants.v+tmp.x;
  self.v.y:=what.y*(self.owner^ as TUniverse).constants.v+tmp.y;
  self.v.z:=what.z*(self.owner^ as TUniverse).constants.v+tmp.z;
end;

{ a jeji precteni}
procedure TPlanet.getvelocity(var what: vektor);
var ix, iy, iz: variant;
begin
  self.GetVelocity(ix,iy,iz);
  what.x:=ix/(self.owner^ as TUniverse).constants.v;
  what.y:=iy/(self.owner^ as TUniverse).constants.v;
  what.z:=iz/(self.owner^ as TUniverse).constants.v;
end;

procedure TPlanet.getvelocity(var ix: variant; var iy: variant; var iz: variant);
var tmp: vektor;
begin
  tmp.null;
  if (self.owner^ as TUniverse).Sit<>NIL then
    tmp:=(self.owner^ as TUniverse).Sit^.v;
  ix:=self.v.x/(self.owner^ as TUniverse).constants.v-tmp.x;
  iy:=self.v.y/(self.owner^ as TUniverse).constants.v-tmp.y;
  iz:=self.v.z/(self.owner^ as TUniverse).constants.v-tmp.z;
end;

{co se stane, kdyz uplyne nejakej cas}
procedure TPlanet.addtime(time: double);
begin
  self.pos:=self.pos.add(self.v.multipliedby(time));
  if self.m<>0 then
    self.v:=self.v.add(self.f.multipliedby((1/self.m)*time));
end;

{jak se vzajemne ovlivnuji dve planety}
procedure TPlanet.Iteraguj(skym: PPlanet);
var r, df: vektor;
begin
  r:=skym.pos.add(self.pos.negative); {vzdalenost planet}
  if r.getnorm>(self.r+skym.r) then
  begin
    df:=r.setnorm((self.owner^ as TUniverse).constants.h*self.m*skym.m/(r.getnorm*r.getnorm));   {pusobici sila}
    self.f:=self.f.add(df);
    skym.f:=skym.f.add(df.negative);
  end;
end;

end.
